import os
import threading
import time
import traceback
from typing import TYPE_CHECKING, Any, List

from ._core import execute_workflow_by_id
from ._error import DBOSWorkflowFunctionNotFoundError
from ._sys_db import GetPendingWorkflowsOutput

if TYPE_CHECKING:
    from ._dbos import DBOS, WorkflowHandle


def startup_recovery_thread(
    dbos: "DBOS", pending_workflows: List[GetPendingWorkflowsOutput]
) -> None:
    """Attempt to recover local pending workflows on startup using a background thread."""
    stop_event = threading.Event()
    dbos.stop_events.append(stop_event)
    while not stop_event.is_set() and len(pending_workflows) > 0:
        try:
            for pending_workflow in list(pending_workflows):
                if (
                    pending_workflow.queue_name
                    and pending_workflow.queue_name != "_dbos_internal_queue"
                ):
                    dbos._sys_db.clear_queue_assignment(pending_workflow.workflow_uuid)
                    continue
                execute_workflow_by_id(dbos, pending_workflow.workflow_uuid)
                pending_workflows.remove(pending_workflow)
        except DBOSWorkflowFunctionNotFoundError:
            time.sleep(1)
        except Exception as e:
            dbos.logger.error(
                f"Exception encountered when recovering workflows: {traceback.format_exc()}"
            )
            raise e


def recover_pending_workflows(
    dbos: "DBOS", executor_ids: List[str] = ["local"]
) -> List["WorkflowHandle[Any]"]:
    workflow_handles: List["WorkflowHandle[Any]"] = []
    for executor_id in executor_ids:
        if executor_id == "local" and os.environ.get("DBOS__VMID"):
            dbos.logger.debug(
                f"Skip local recovery because it's running in a VM: {os.environ.get('DBOS__VMID')}"
            )
        dbos.logger.debug(f"Recovering pending workflows for executor: {executor_id}")
        pending_workflows = dbos._sys_db.get_pending_workflows(executor_id)
        for pending_workflow in pending_workflows:
            if (
                pending_workflow.queue_name
                and pending_workflow.queue_name != "_dbos_internal_queue"
            ):
                try:
                    dbos._sys_db.clear_queue_assignment(pending_workflow.workflow_uuid)
                    workflow_handles.append(
                        dbos.retrieve_workflow(pending_workflow.workflow_uuid)
                    )
                except Exception as e:
                    dbos.logger.error(e)
            else:
                workflow_handles.append(
                    execute_workflow_by_id(dbos, pending_workflow.workflow_uuid)
                )

    dbos.logger.info("Recovered pending workflows")
    return workflow_handles



import threading
import time
from datetime import datetime, timedelta, timezone

# Public API
from dbos import DBOS, ConfigFile, Queue, WorkflowStatusString, _workflow_commands


def test_list_workflow(dbos: DBOS, config: ConfigFile) -> None:
    print("Testing list_workflow")

    @DBOS.workflow()
    def simple_workflow() -> None:
        print("Executed Simple workflow")
        return

    # run the workflow
    simple_workflow()
    time.sleep(1)  # wait for the workflow to complete
    # get the workflow list
    output = _workflow_commands.list_workflows(
        config, 10, None, None, None, None, False, None, None
    )
    assert len(output) == 1, f"Expected list length to be 1, but got {len(output)}"
    assert output[0] != None, "Expected output to be not None"
    if output[0] != None:
        assert output[0].workflowUUID.strip(), "field_name is an empty string"


def test_list_workflow_limit(dbos: DBOS, config: ConfigFile) -> None:
    print("Testing list_workflow")

    @DBOS.workflow()
    def simple_workflow() -> None:
        print("Executed Simple workflow")
        return

    # run the workflow
    simple_workflow()
    simple_workflow()
    simple_workflow()
    time.sleep(1)  # wait for the workflow to complete
    # get the workflow list
    output = _workflow_commands.list_workflows(
        config, 2, None, None, None, None, False, None, None
    )
    assert len(output) == 2, f"Expected list length to be 1, but got {len(output)}"


def test_list_workflow_status_name(dbos: DBOS, config: ConfigFile) -> None:
    print("Testing list_workflow")

    @DBOS.workflow()
    def simple_workflow() -> None:
        print("Executed Simple workflow")
        return

    # run the workflow
    simple_workflow()
    time.sleep(1)  # wait for the workflow to complete
    # get the workflow list
    output = _workflow_commands.list_workflows(
        config, 10, None, None, None, "PENDING", False, None, None
    )
    assert len(output) == 0, f"Expected list length to be 0, but got {len(output)}"

    output = _workflow_commands.list_workflows(
        config, 10, None, None, None, "SUCCESS", False, None, None
    )
    assert len(output) == 1, f"Expected list length to be 1, but got {len(output)}"

    output = _workflow_commands.list_workflows(
        config, 10, None, None, None, None, False, None, "no"
    )
    assert len(output) == 0, f"Expected list length to be 0, but got {len(output)}"

    output = _workflow_commands.list_workflows(
        config, 10, None, None, None, None, False, None, simple_workflow.__qualname__
    )
    assert len(output) == 1, f"Expected list length to be 1, but got {len(output)}"


def test_list_workflow_start_end_times(dbos: DBOS, config: ConfigFile) -> None:
    print("Testing list_workflow")

    @DBOS.workflow()
    def simple_workflow() -> None:
        print("Executed Simple workflow")
        return

    now = datetime.now()
    starttime = (now - timedelta(seconds=20)).isoformat()
    print(starttime)

    # run the workflow
    simple_workflow()
    time.sleep(1)  # wait for the workflow to complete
    # get the workflow list
    endtime = datetime.now().isoformat()
    print(endtime)

    output = _workflow_commands.list_workflows(
        config, 10, None, starttime, endtime, None, False, None, None
    )
    assert len(output) == 1, f"Expected list length to be 1, but got {len(output)}"

    newstarttime = (now - timedelta(seconds=30)).isoformat()
    newendtime = starttime

    output = _workflow_commands.list_workflows(
        config, 10, None, newstarttime, newendtime, None, False, None, None
    )
    assert len(output) == 0, f"Expected list length to be 0, but got {len(output)}"


def test_list_workflow_end_times_positive(dbos: DBOS, config: ConfigFile) -> None:
    print("Testing list_workflow")

    @DBOS.workflow()
    def simple_workflow() -> None:
        print("Executed Simple workflow")
        return

    now = datetime.now()

    time_0 = (now - timedelta(seconds=40)).isoformat()

    time_1 = (now - timedelta(seconds=20)).isoformat()

    # run the workflow
    simple_workflow()
    time.sleep(1)  # wait for the workflow to complete
    # get the workflow list
    time_2 = datetime.now().isoformat()

    simple_workflow()
    time.sleep(1)  # wait for the workflow to complete
    # get the workflow list
    time_3 = datetime.now().isoformat()

    output = _workflow_commands.list_workflows(
        config, 10, None, time_0, time_1, None, False, None, None
    )

    assert len(output) == 0, f"Expected list length to be 0, but got {len(output)}"

    output = _workflow_commands.list_workflows(
        config, 10, None, time_1, time_2, None, False, None, None
    )

    assert len(output) == 1, f"Expected list length to be 1, but got {len(output)}"

    output = _workflow_commands.list_workflows(
        config, 10, None, time_1, time_3, None, False, None, None
    )
    assert len(output) == 2, f"Expected list length to be 2, but got {len(output)}"


def test_get_workflow(dbos: DBOS, config: ConfigFile) -> None:
    print("Testing get_workflow")

    @DBOS.workflow()
    def simple_workflow() -> None:
        print("Executed Simple workflow")
        return

    # run the workflow
    simple_workflow()
    time.sleep(1)  # wait for the workflow to complete
    # get the workflow list
    output = _workflow_commands.list_workflows(
        config, 10, None, None, None, None, False, None, None
    )
    assert len(output) == 1, f"Expected list length to be 1, but got {len(output)}"

    assert output[0] != None, "Expected output to be not None"

    wfUuid = output[0].workflowUUID

    info = _workflow_commands.get_workflow(config, wfUuid, True)
    assert info is not None, "Expected output to be not None"

    if info is not None:
        assert info.workflowUUID == wfUuid, f"Expected workflow_uuid to be {wfUuid}"


def test_cancel_workflow(dbos: DBOS, config: ConfigFile) -> None:
    print("Testing get_workflow")

    @DBOS.workflow()
    def simple_workflow() -> None:
        time.sleep(3)
        print("Executed Simple workflow")
        return

    # run the workflow
    simple_workflow()
    # get the workflow list
    output = _workflow_commands.list_workflows(
        config, 10, None, None, None, None, False, None, None
    )
    # assert len(output) == 1, f"Expected list length to be 1, but got {len(output)}"

    print(output[0])
    assert output[0] != None, "Expected output to be not None"
    wfUuid = output[0].workflowUUID

    _workflow_commands.cancel_workflow(config, wfUuid)

    info = _workflow_commands.get_workflow(config, wfUuid, True)
    assert info is not None, "Expected info to be not None"
    if info is not None:
        assert info.status == "CANCELLED", f"Expected status to be CANCELLED"


def test_queued_workflows(dbos: DBOS, config: ConfigFile) -> None:
    queued_steps = 5
    step_events = [threading.Event() for _ in range(queued_steps)]
    event = threading.Event()
    queue = Queue("test_queue")

    @DBOS.workflow()
    def test_workflow() -> list[int]:
        handles = []
        for i in range(queued_steps):
            h = queue.enqueue(blocking_step, i)
            handles.append(h)
        return [h.get_result() for h in handles]

    @DBOS.step()
    def blocking_step(i: int) -> int:
        step_events[i].set()
        event.wait()
        return i

    # The workflow enqueues blocking steps, wait for all to start
    handle = DBOS.start_workflow(test_workflow)
    for e in step_events:
        e.wait()

    # Verify all blocking steps are enqueued and have the right data
    workflows = _workflow_commands.list_queued_workflows(config)
    assert len(workflows) == queued_steps
    for i, workflow in enumerate(workflows):
        assert workflow.status == WorkflowStatusString.PENDING.value
        assert workflow.queue_name == queue.name
        assert workflow.input is not None
        # Verify oldest queue entries appear first
        assert workflow.input["args"][0] == i
        assert workflow.output is None
        assert workflow.error is None
        assert "blocking_step" in workflow.workflowName

    # Test every filter
    workflows = _workflow_commands.list_queued_workflows(
        config, status=WorkflowStatusString.PENDING.value
    )
    assert len(workflows) == queued_steps
    workflows = _workflow_commands.list_queued_workflows(
        config, status=WorkflowStatusString.ENQUEUED.value
    )
    assert len(workflows) == 0
    workflows = _workflow_commands.list_queued_workflows(config, queue_name=queue.name)
    assert len(workflows) == queued_steps
    workflows = _workflow_commands.list_queued_workflows(config, queue_name="no")
    assert len(workflows) == 0
    workflows = _workflow_commands.list_queued_workflows(
        config, name=f"<temp>.{blocking_step.__qualname__}"
    )
    assert len(workflows) == queued_steps
    workflows = _workflow_commands.list_queued_workflows(config, name="no")
    assert len(workflows) == 0
    now = datetime.now(timezone.utc)
    start_time = (now - timedelta(seconds=10)).isoformat()
    end_time = (now + timedelta(seconds=10)).isoformat()
    workflows = _workflow_commands.list_queued_workflows(
        config, start_time=start_time, end_time=end_time
    )
    assert len(workflows) == queued_steps
    workflows = _workflow_commands.list_queued_workflows(
        config, start_time=now.isoformat(), end_time=end_time
    )
    assert len(workflows) == 0

    # Confirm the workflow finishes and nothing is enqueued afterwards
    event.set()
    assert handle.get_result() == [0, 1, 2, 3, 4]
    workflows = _workflow_commands.list_queued_workflows(config)
    assert len(workflows) == 0



from typing import List, Optional, cast

import typer

from . import _serialization
from ._dbos_config import ConfigFile
from ._sys_db import (
    GetQueuedWorkflowsInput,
    GetWorkflowsInput,
    GetWorkflowsOutput,
    SystemDatabase,
    WorkflowStatuses,
)


class WorkflowInformation:
    workflowUUID: str
    status: WorkflowStatuses
    workflowName: str
    workflowClassName: Optional[str]
    workflowConfigName: Optional[str]
    input: Optional[_serialization.WorkflowInputs]  # JSON (jsonpickle)
    output: Optional[str] = None  # JSON (jsonpickle)
    error: Optional[str] = None  # JSON (jsonpickle)
    executor_id: Optional[str]
    app_version: Optional[str]
    app_id: Optional[str]
    request: Optional[str]  # JSON (jsonpickle)
    recovery_attempts: Optional[int]
    authenticated_user: Optional[str]
    assumed_role: Optional[str]
    authenticated_roles: Optional[str]  # JSON list of roles.
    queue_name: Optional[str]


def list_workflows(
    config: ConfigFile,
    limit: int,
    user: Optional[str],
    starttime: Optional[str],
    endtime: Optional[str],
    status: Optional[str],
    request: bool,
    appversion: Optional[str],
    name: Optional[str],
) -> List[WorkflowInformation]:
    try:
        sys_db = SystemDatabase(config)

        input = GetWorkflowsInput()
        input.authenticated_user = user
        input.start_time = starttime
        input.end_time = endtime
        if status is not None:
            input.status = cast(WorkflowStatuses, status)
        input.application_version = appversion
        input.limit = limit
        input.name = name

        output: GetWorkflowsOutput = sys_db.get_workflows(input)
        infos: List[WorkflowInformation] = []
        for workflow_id in output.workflow_uuids:
            info = _get_workflow_info(
                sys_db, workflow_id, request
            )  # Call the method for each ID
            if info is not None:
                infos.append(info)

        return infos
    except Exception as e:
        typer.echo(f"Error listing workflows: {e}")
        return []
    finally:
        if sys_db:
            sys_db.destroy()


def list_queued_workflows(
    config: ConfigFile,
    limit: Optional[int] = None,
    start_time: Optional[str] = None,
    end_time: Optional[str] = None,
    queue_name: Optional[str] = None,
    status: Optional[str] = None,
    name: Optional[str] = None,
    request: bool = False,
) -> List[WorkflowInformation]:
    try:
        sys_db = SystemDatabase(config)
        input: GetQueuedWorkflowsInput = {
            "queue_name": queue_name,
            "start_time": start_time,
            "end_time": end_time,
            "status": status,
            "limit": limit,
            "name": name,
        }
        output: GetWorkflowsOutput = sys_db.get_queued_workflows(input)
        infos: List[WorkflowInformation] = []
        for workflow_id in output.workflow_uuids:
            info = _get_workflow_info(
                sys_db, workflow_id, request
            )  # Call the method for each ID
            if info is not None:
                infos.append(info)
        return infos
    except Exception as e:
        typer.echo(f"Error listing workflows: {e}")
        return []
    finally:
        if sys_db:
            sys_db.destroy()


def get_workflow(
    config: ConfigFile, uuid: str, request: bool
) -> Optional[WorkflowInformation]:
    try:
        sys_db = SystemDatabase(config)
        info = _get_workflow_info(sys_db, uuid, request)
        return info
    except Exception as e:
        typer.echo(f"Error getting workflow: {e}")
        return None
    finally:
        if sys_db:
            sys_db.destroy()


def cancel_workflow(config: ConfigFile, uuid: str) -> None:
    try:
        sys_db = SystemDatabase(config)
        sys_db.cancel_workflow(uuid)
    except Exception as e:
        typer.echo(f"Failed to connect to DBOS system database: {e}")
        raise e
    finally:
        if sys_db:
            sys_db.destroy()


def _get_workflow_info(
    sys_db: SystemDatabase, workflowUUID: str, getRequest: bool
) -> Optional[WorkflowInformation]:

    info = sys_db.get_workflow_status(workflowUUID)
    if info is None:
        return None

    winfo = WorkflowInformation()

    winfo.workflowUUID = workflowUUID
    winfo.status = info["status"]
    winfo.workflowName = info["name"]
    winfo.workflowClassName = info["class_name"]
    winfo.workflowConfigName = info["config_name"]
    winfo.executor_id = info["executor_id"]
    winfo.app_version = info["app_version"]
    winfo.app_id = info["app_id"]
    winfo.recovery_attempts = info["recovery_attempts"]
    winfo.authenticated_user = info["authenticated_user"]
    winfo.assumed_role = info["assumed_role"]
    winfo.authenticated_roles = info["authenticated_roles"]
    winfo.queue_name = info["queue_name"]

    # no input field
    input_data = sys_db.get_workflow_inputs(workflowUUID)
    if input_data is not None:
        winfo.input = input_data

    if info.get("status") == "SUCCESS":
        result = sys_db.await_workflow_result(workflowUUID)
        winfo.output = result
    elif info.get("status") == "ERROR":
        try:
            sys_db.await_workflow_result(workflowUUID)
        except Exception as e:
            winfo.error = str(e)

    if not getRequest:
        winfo.request = None

    return winfo
